'''

Speech-to-text


'''

from thespian.actors import Actor
from pygls.server import LanguageServer
from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Position,
    TextDocumentIdentifier,
)
import logging
from threading import Thread, Event
from queue import Queue, Empty
import speech_recognition as sr
import re
import numpy as np
import time
import argparse
import threading
from functools import partial

from uniteai.common import mk_logger, find_block, get_nested
from uniteai.edit import BlockJob, cleanup_block, init_block

START_TAG = ':START_TRANSCRIPTION:'
END_TAG = ':END_TRANSCRIPTION:'
NAME = 'transcription'

# A custom logger for just this feature. You can tune the log level to turn
# on/off just this feature's logs.
log_level = logging.DEBUG
log = mk_logger(NAME, log_level)


##################################################
# SpeechRecognition

class SpeechRecognition:
    def __init__(self,
                 model_type,

                 # whisper
                 model_path,
                 model_size,
                 volume_threshold):
        self.model_type = model_type
        self.model_path = model_path
        self.model_size = model_size

        # Recognizer
        self.r = sr.Recognizer()
        self.mic = sr.Microphone()
        self.sample_rate = None
        self.sample_width = None
        self.r.energy_threshold = volume_threshold
        self.r.dynamic_energy_threshold = False

    def recognize(self, audio):
        log.debug(f'MODEL_TYPE: {self.model_type}')
        log.debug(f'AUDIO: {audio}')
        if self.model_type == 'vosk':
            return self.r.recognize_whisper(
                audio,
                load_options=dict(
                    device='cuda:0',
                    download_root=self.model_path
                ), language='en')

        if self.model_type == 'whisper':
            return self.r.recognize_whisper(
                audio,
                model=self.model_size,
                load_options=dict(
                    device='cuda:0',
                    download_root=self.model_path
                ), language='english')

    def _warmup(self):
        ''' Warm up, intended for a separate thread. '''

        # Get some mic params
        with self.mic as source:
            self.sample_rate = source.SAMPLE_RATE
            self.sample_width = source.SAMPLE_WIDTH

        # Get model into memory
        empty_audio = sr.AudioData(np.zeros(10), sample_rate=1, sample_width=1)
        self.recognize(empty_audio)
        log.info(f'Warmed up. sample_rate={self.sample_rate}, sample_width={self.sample_width}')

    def warmup(self):
        '''Load whisper model into memory.'''
        log.info('Warming up transcription model in separate thread')
        warmup_thread = Thread(target=self._warmup)
        warmup_thread.daemon = True
        warmup_thread.start()

    def listen_(self,
                queue: Queue,
                should_stop: Event):
        with sr.Microphone() as s:
            while not should_stop.is_set():
                buf = s.stream.read(s.CHUNK)
                queue.put(buf)

    def transcription_(self,
                       audio_queue,
                       transcription_callback,
                       finished_callback,
                       should_stop):
        audios = []
        while not should_stop.is_set():
            try:
                # non-blocking, to more frequently allow the
                # `stop_transcription` signal to end this thread.
                buffer = audio_queue.get(False)

                # TODO: can we more intelligently separate silence from speech?
                # energy = audioop.rms(buffer, self.sample_width)
                audios.append(buffer)
                try:
                    while True:
                        buffer = audio_queue.get(False)
                        audios.append(buffer)
                except Empty:
                    pass

                log.debug(f'len audio: {len(audios)}')

            except Empty:
                time.sleep(0.2)
                continue

            try:
                audio = sr.audio.AudioData(
                    b''.join(audios),
                    self.sample_rate,
                    self.sample_width
                )
                # Debug audio: The audio gets sliced up regularly, how does it
                #              sound when stitched back?
                if log_level == logging.DEBUG:
                    with open("debug_transcription.wav", "wb") as output_file:
                        output_file.write(audio.get_wav_data())

                # breakout if needed
                if should_stop.is_set():
                    break

                # Speech-to-text
                x = self.recognize(audio)

                # Nothing recognized
                if not x:
                    continue

                x = x.strip()
                if filter_out(x):
                    continue

                # breakout if needed
                if should_stop.is_set():
                    break

                transcription_callback(x)

            except sr.UnknownValueError:
                log.debug("ERROR: could not understand audio")
            audio_queue.task_done()

        finished_callback()
        log.debug('DONE TRANSCRIBING')

    def go(self,
           transcription_callback,
           finished_callback):
        audio_queue = Queue()
        should_stop = Event()

        # Listener Thread
        l_thread = threading.Thread(
            target=self.listen_,
            args=(audio_queue, should_stop))
        l_thread.daemon = True
        l_thread.start()

        # Transcription Thread
        t_thread = threading.Thread(
            target=self.transcription_,
            args=(audio_queue,
                  transcription_callback,
                  finished_callback,
                  should_stop))
        t_thread.daemon = True
        t_thread.start()

        def stop_fn():
            log.debug('stop_fn called')
            should_stop.set()
            l_thread.join()
            t_thread.join()

        return stop_fn


##################################################
# Actor

class TranscriptionActor(Actor):
    def __init__(self):
        self.is_running = Event()
        self.tags = [START_TAG, END_TAG]
        self.speech_recognition = None  # set during initialization

        # set during set_config/start
        self.model_path = None
        self.model_size = None
        self.volume_threshold = None
        self.stop_fn = lambda: None

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        edits = msg.get('edits')
        tw_set = self.is_running.is_set()
        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
transcription_worker is running={tw_set}

EDITS STATE:
job_thread alive: {edits.job_thread.is_alive() if edits and edits.job_thread else "NOT STARTED"}
%%%%%%%%%%
''')

        ##########
        # Start
        if command == 'start':
            uri = msg.get('uri')
            cursor_pos = msg.get('cursor_pos')
            doc = msg.get('doc')

            # check if block already exists
            start_ixs, end_ixs = find_block(START_TAG,
                                            END_TAG,
                                            doc)
            # make block
            if not (start_ixs and end_ixs):
                init_block(NAME, self.tags, uri, cursor_pos, edits)

            self.start(uri, cursor_pos, edits)

        ##########
        # Stop
        elif command == 'stop':
            self.stop()

        ##########
        # Set Config
        elif command == 'set_config':
            config = msg.get('config')
            self.model_type = config.transcription_model_type
            self.volume_threshold = config.transcription_volume_threshold
            self.model_path = config.transcription_model_path

            # Vosk
            if self.model_type == 'vosk':
                pass

            # Whisper
            if self.model_type == 'whisper':
                self.model_size = config.transcription_model_size

        elif command == 'initialize':
            log.debug(f'INIT TYPE: {self.model_type}')
            self.speech_recognition = SpeechRecognition(
                self.model_type,

                # whisper
                self.model_path,
                self.model_size,
                self.volume_threshold
            )

            # load the model into GPU
            self.speech_recognition.warmup()

    def transcription_callback(self, edits, uri, text):
        # Add space to respect next loop of transcription
        log.debug(f'TRANSCRIBED: {text}')
        job = BlockJob(
            uri=uri,
            start_tag=START_TAG,
            end_tag=END_TAG,
            text=f'\n{text}\n',
            strict=False,
        )
        edits.add_job(NAME, job)

    def finished_callback(self, edits, uri):
        log.debug(f'FINISHED CALLBACK: {uri}')
        cleanup_block(NAME, [START_TAG, END_TAG], uri, edits)

    def start(self, uri, cursor_pos, edits):
        if self.is_running.is_set():
            log.info('WARN: ON_START_BUT_RUNNING.')
            return False
        self.stop_fn = self.speech_recognition.go(
            partial(self.transcription_callback, edits, uri),
            partial(self.finished_callback, edits, uri))
        self.is_running.set()
        log.debug('START CAN RETURN')

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running.is_set():
            log.info('WARN: ON_STOP_BUT_STOPPED')
            return False
        self.stop_fn()
        self.is_running.clear()
        self.stop_fn = lambda: None
        log.debug('FINALLY STOPPED')


##########
# Util

def filter_alphanum(x: str) -> str:
    ''' keep only alphanum, not even spaces nor punctuation. '''
    return re.sub(r'\W+', '', x)


filter_list = [
    filter_alphanum(x)  # removes spaces
    for x in
    [  # quirks of Whisper
        '',
        'bye',
        'you',
        'thank you',
        'thanks for watching',
    ]
]


def filter_out(x: str) -> bool:
    x = filter_alphanum(x)
    return x.strip().lower() in filter_list


def code_action_transcribe(params: CodeActionParams):
    '''Trigger a ChatGPT response. A code action calls a command, which is set
    up below to `tell` the actor to start streaming a response. '''
    text_document = params.text_document
    range = params.range  # lsp spec only provides `Range`
    cursor_pos = range.end
    return CodeAction(
        title='Transcribe',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='Transcribe',
            command='command.transcribe',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, cursor_pos]
        )
    )


##################################################

def configure(config_yaml):
    parser = argparse.ArgumentParser()
    parser.add_argument('--transcription_model_type', default=get_nested(config_yaml, ['transcription', 'model_type']))
    parser.add_argument('--transcription_model_path', default=get_nested(config_yaml, ['transcription', 'model_path']))
    parser.add_argument('--transcription_volume_threshold', default=get_nested(config_yaml, ['transcription', 'volume_threshold']))

    # whisper
    parser.add_argument('--transcription_model_size', default=get_nested(config_yaml, ['transcription', 'model_size']))

    # bc this is only concerned with transcription params, do not error if extra
    # params are sent via cli.
    args, _ = parser.parse_known_args()
    return args


def initialize(config: argparse.Namespace, server):
    # Actor
    server.add_actor(NAME, TranscriptionActor)
    server.tell_actor(NAME, {
        'command': 'set_config',
        'config': config
    })
    server.tell_actor(NAME, {
        'command': 'initialize'
    })

    # CodeActions
    server.add_code_action(code_action_transcribe)

    # Modify Server
    @server.thread()
    @server.command('command.transcribe')
    def transcribe_stream(ls: LanguageServer, args):
        if len(args) != 2:
            log.error(f'command.transcribe: Wrong arguments, received: {args}')
        # Prepare args
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        cursor_pos = ls.converter.structure(args[1], Position)
        actor_args = {
            'command': 'start',
            'uri': uri,
            'doc': doc.source,
            'edits': ls.edits,
            'cursor_pos': cursor_pos,
        }
        ls.tell_actor(NAME, actor_args)
        return {'status': 'success'}



'''

An LSP server that connects to the LLM server and APIs for doing the brainy
stuff.


USAGE:
    python lsp_server.py

'''


import sys
import logging
import uniteai.server
import uniteai.config as config
import importlib


##########
# Logging
#
# Change the basicConfig level to be the generous DEBUG, quiet the libraries,
# and allow custom loggers in modules to have their own debug levels. This
# helps especially with debugging weird concurrency quirks, ie allowing
# different processes to report things as needed, in a way that can be easily
# turned on and off per object of your attention.
#
# NOTE: Thespian has a domineering logging methodology. To customize their
#       formatter, see: `thespian.system.simpleSystemBase`.
#
#       Also, https://github.com/thespianpy/Thespian/issues/73
#
# NOTE: VSCode thinks the server is erroring if it logs to stdout from the main
#       thread, but spawned threads don't seem to have the same effect.

logging.basicConfig(
    filename='log_file.log', # TODO: feature loggers still seem to report on stdout
    # stream=sys.stdout,
    level=logging.DEBUG,
)

# Quiet the libs a little
logging.getLogger('pygls.feature_manager').setLevel(logging.WARN)
logging.getLogger('pygls.protocol').setLevel(logging.WARN)
logging.getLogger('Thespian').setLevel(logging.WARN)
logging.getLogger('asyncio').setLevel(logging.WARN)


def load_module(module_name, config_yaml, server):
    ''' Useful for loading just the modules referenced in the config. '''
    module = importlib.import_module(module_name)
    if hasattr(module, 'configure'):
        args = module.configure(config_yaml)

    if hasattr(module, 'initialize'):
        module.initialize(args, server)


def main():
    # First pass at configuration. Further passes will pick up config
    # per-feature.
    args, config_yaml, parser = config.get_args()
    server = uniteai.server.initialize()

    for module_name in args.modules:
        load_module(module_name, config_yaml, server)

    if args.tcp:
        print(f'Starting LSP on port: {args.lsp_port}')
        server.start_tcp(host='localhost', port=args.lsp_port)
    elif args.stdio:
        print(f'Starting on STDIO')
        server.start_io()


if __name__ == '__main__':
    main()



'''

Start building your feature off this example!

This example inserts auto-incrementing numbers.

'''


from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Position,
    TextDocumentIdentifier,
    WorkspaceEdit,
)
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from thespian.actors import Actor
import argparse
import logging
import time

from uniteai.edit import init_block, cleanup_block, BlockJob
from uniteai.common import find_block, mk_logger, get_nested
from uniteai.server import Server


START_TAG = ':START_EXAMPLE:'
END_TAG = ':END_EXAMPLE:'
NAME = 'example'

# A custom logger for just this feature. You can tune the log level to turn
# on/off just this feature's logs.
log = mk_logger(NAME, logging.WARN)


class ExampleActor(Actor):
    def __init__(self):
        self.is_running = False
        self.executor = ThreadPoolExecutor(max_workers=3)
        self.current_future = None
        self.should_stop = Event()
        self.tags = [START_TAG, END_TAG]

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        doc = msg.get('doc')
        edits = msg.get('edits')

        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
is_running: {self.is_running}
locked: {self.should_stop.is_set()}
future: {self.current_future}

EDITS STATE:
job_thread alive: {edits.job_thread.is_alive() if edits and edits.job_thread else "NOT STARTED"}
%%%%%%%%%%''')

        ##########
        # Start
        if command == 'start':
            uri = msg.get('uri')
            cursor_pos = msg.get('cursor_pos')
            engine = msg.get('engine')
            max_length = msg.get('max_length')
            edits = msg.get('edits')

            # check if block already exists
            start_ixs, end_ixs = find_block(START_TAG,
                                            END_TAG,
                                            doc)

            if not (start_ixs and end_ixs):
                init_block(NAME, self.tags, uri, cursor_pos, edits)

            self.start(uri, cursor_pos, engine, max_length, edits)

        ##########
        # Stop
        elif command == 'stop':
            self.stop()

        ##########
        # Set Config
        elif command == 'set_config':
            config = msg['config']
            self.start_digit = config.example_start_digit
            self.end_digit = config.example_end_digit
            self.delay = config.example_delay

    def start(self, uri, cursor_pos, engine, max_length, edits):
        if self.is_running:
            log.info('WARN: ON_START_BUT_RUNNING')
            return
        log.debug('ACTOR START')

        self.is_running = True
        self.should_stop.clear()
        self.current_future = self.executor.submit(
            self.stream_fn, uri, self.should_stop, edits
        )
        log.debug('START CAN RETURN')

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running:
            log.info('WARN: ON_STOP_BUT_STOPPED')

        self.should_stop.set()

        if self.current_future:
            self.current_future.result()  # block, wait to finish
            self.current_future = None
        log.debug('FINALLY STOPPED')

    def stream_fn(self, uri, stop_event, edits):
        log.debug('START: EXAMPLE_STREAM_FN')
        try:
            # Stream the results to LSP Client
            running_text = ''
            for x in range(self.start_digit, self.end_digit+1):
                # For breaking out early
                if stop_event.is_set():
                    log.debug('STREAM_FN received STOP EVENT')
                    break

                running_text += f'{x} '
                job = BlockJob(
                    uri=uri,
                    start_tag=START_TAG,
                    end_tag=END_TAG,
                    text=f'\n{running_text}\n',
                    # "non-strict" means that if this job doesn't get applied
                    # successfully, it can be dropped. This is useful when
                    # streaming.
                    strict=False,
                )
                edits.add_job(NAME, job)
                time.sleep(self.delay)

            # Streaming is done, and those added jobs were all
            # non-strict. Let's make sure to have one final strict
            # job. Streaming jobs are ok to be dropped, but we need to make
            # sure it does finalize, eg before a strict delete-tags job is
            # added.
            job = BlockJob(
                uri=uri,
                start_tag=START_TAG,
                end_tag=END_TAG,
                text=f'\n{running_text}\n',
                strict=True,
            )
            edits.add_job(NAME, job)

        except Exception as e:
            log.error(f'Error: ExapleActor, {e}')

        # Cleanup
        log.debug('CLEANING UP')
        cleanup_block(NAME, self.tags, uri, edits)
        self.is_running = False
        self.current_future = None
        self.should_stop.clear()


def code_action_example(start_digit: int,
                        end_digit: int,
                        delay: int,
                        params: CodeActionParams):
    ''' A code_action triggers a command, which sends a message to the Actor,
    to handle it. '''
    text_document = params.text_document
    # position of the highlighted region in the client's editor
    range = params.range  # lsp spec only provides `Range`
    cursor_pos = range.end
    return CodeAction(
        title='Example Counter',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='Example Counter',
            command='command.exampleCounter',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, cursor_pos, start_digit, end_digit, delay]
        )
    )


##################################################
# Setup
#
# NOTE: In `.uniteai.yml`, just add `uniteai.example` under `modules`, and this
#       will automatically get built into the server at runtime.
#

def configure(config_yaml):
    parser = argparse.ArgumentParser()
    parser.add_argument('--example_start_digit', default=get_nested(config_yaml, ['example', 'start_digit']))
    parser.add_argument('--example_end_digit', default=get_nested(config_yaml, ['example', 'end_digit']))
    parser.add_argument('--example_delay', default=get_nested(config_yaml, ['example', 'delay']))

    # These get picked up as `config` in `initialize`

    # bc this is only concerned with example params, do not error if extra
    # params are sent via cli.
    args, _ = parser.parse_known_args()
    return args


def initialize(config, server):
    # Config
    start_digit = config.example_start_digit
    end_digit = config.example_end_digit
    delay = config.example_delay

    # Actor
    server.add_actor(NAME, ExampleActor)

    # Initialize configuration in Actor
    server.tell_actor(NAME, {
        'command': 'set_config',
        'config': config,
    })

    # CodeActions
    server.add_code_action(
        lambda params:
        code_action_example(start_digit, end_digit, delay, params))

    # Modify Server
    @server.thread()
    @server.command('command.exampleCounter')
    def example_counter(ls: Server, args):
        if len(args) != 2:
            log.error(f'command.exampleCounter: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        cursor_pos = ls.converter.structure(args[1], Position)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Send a message to start the stream
        actor_args = {
            'command': 'start',
            'uri': uri,
            'cursor_pos': cursor_pos,
            'start_digit': start_digit,
            'end_digit': end_digit,
            'delay': delay,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)

        # Return null-edit immediately (the rest will stream)
        return WorkspaceEdit()



'''

Configuration alternatively comes from a config file, or is overridden by a
command line argument.


ORDER OF PRIORITIES

1.) CLI arguments
2.) ./.uniteai.yml
3.) if not found, then read  ~/.uniteai.yml

'''

import argparse
import yaml
import os
import sys
import shutil
import logging
from uniteai.common import mk_logger

log = mk_logger('CONFIG', logging.WARN)

CONFIG_PATHS = [
    '.uniteai.yml',
    '.uniteai.yaml',
    os.path.expanduser('~/.uniteai.yaml'),
    os.path.expanduser('~/.uniteai.yml'),
]


def handle_missing_config():
    '''Possibly create a default config, and then exit.'''
    log.error('No config file found!')
    ans = input('Would you like this process to copy the default `.uniteai.yml.example` config file into the current directory? (y)es / (n)o')

    if ans.lower() in ['y', 'yes']:
        log.info('''
Copying `.uniateai.yml.example` to `.uniteai.yml`
Please review it before running the LSP again.
It requires secrets (eg OpenAI Key) so you may prefer to locate it at `~/.uniteai.yml`.'''.strip())

        # New path
        config_path = '.uniteai.yml'

        # Example path
        example_config_path = '.uniteai.yml.example'

        if not os.path.exists(config_path):
            shutil.copyfile(example_config_path, config_path)
    else:
        log.error('''Please manually copy and update the file `.uniateai.yml.example` from https://github.com/freckletonj/uniteai.''')
    sys.exit(1)


def load_config(file_paths=CONFIG_PATHS):
    ''' Return first config file that exists. '''
    for file_path in file_paths:
        if os.path.exists(file_path):
            log.info(f'Reading configuration from: {file_path}')
            with open(file_path, 'r') as f:
                return yaml.safe_load(f)

def get_args():
    ''' This first pass will learn generic LSP-related config, and what further
    modules need to be loaded. Those modules will be able to specify their own
    configuration, which will be gathered in a second round of config parsing.
    '''
    # Load config file
    config_yaml = load_config(CONFIG_PATHS)

    if config_yaml is None:
        # Will exit after possibly copying a new config
        handle_missing_config()

    # Parse command-line arguments
    parser = argparse.ArgumentParser()

    # LSP-related config
    parser.add_argument('--stdio', action='store_true', default=True)
    parser.add_argument('--tcp', action='store_true')
    parser.add_argument('--lsp_port', default=config_yaml.get('lsp_port', None))
    parser.add_argument('--modules', default=config_yaml.get('modules', None))

    return parser.parse_args(), config_yaml, parser



'''

The Pygls LSP Server, and functions for adding features into it.

'''

from typing import List, Callable
from pygls.server import LanguageServer
from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    TextDocumentIdentifier,

    TEXT_DOCUMENT_DID_OPEN,
    DidOpenTextDocumentParams,
    Diagnostic,
    Range,
    Position,
)
from pygls.protocol import default_converter
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from thespian.actors import ActorSystem
import uniteai.edit as edit
import logging
from uniteai.common import mk_logger

NAME = 'server.py'
log = mk_logger(NAME, logging.DEBUG)

##################################################
# Server

class Server(LanguageServer):
    def __init__(self, name, version):
        super().__init__(name, version)
        self.code_actions = []

        # A ThreadPool for tasks. By throwing long running tasks on here, LSP
        # endpoints can return immediately.
        self.executor = ThreadPoolExecutor(max_workers=3)

        # For converting jsonified things into proper instances. This is
        # necessary when a CodeAction summons a Command, and passes raw JSON to
        # it.
        self.converter = default_converter()

        # Edit-jobs
        edit_job_delay = 0.2
        self.edits = edit.Edits(
            lambda job: edit._attempt_edit_job(self, job), edit_job_delay)
        self.edits.start()

        # Actor System
        self.actor_system = ActorSystem()
        self.actors = {}  # each feature (eg Local LLM) gets its own actor)

    def add_actor(self, name, cls):
        ''' Adds an Actor, and an Edits job queue '''
        self.edits.create_job_queue(name)
        actor = self.actor_system.createActor(
            cls,
            globalName=name
        )
        self.actors[name] = actor
        return actor

    def add_code_action(self, f: Callable[[CodeActionParams], CodeAction]):
        self.code_actions.append(f)

    def tell_actor(self, name, msg):
        self.actor_system.tell(self.actors[name], msg)


def initialize():
    '''
    A Barebones pygls LSP Server.
    '''
    server = Server("uniteai", "0.1.0")

    @server.feature('workspace/didChangeConfiguration')
    def workspace_did_change_configuration(ls: Server, args):
        ''' There's a warning without this. '''
        log.debug(f'workspace/didChangeConfiguration: args={args}')
        return []

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    async def did_open(ls: Server, params: DidOpenTextDocumentParams):
        """Text document did open notification. It appears this does not
        overwrite the default `didOpen` handler in pygls, so we can add extra
        `didOpen` logic here."""
        log.debug(f'Document did open: uri={params.text_document.uri}')

    @server.feature("textDocument/codeAction")
    def code_action(params: CodeActionParams) -> List[CodeAction]:
        ''' Code Actions are eg a list of options you can trigger in your
        client. Included modules can add to `server.code_actions` via
        `server.add_code_action`.  '''
        return [action(params) for action in server.code_actions]

    @server.thread()
    @server.command('command.stop')
    def stop(ls: Server, args):
        '''
        Tell ALL actors to stop.
        '''
        if len(args) != 1:
            log.error(f'command.stop: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Send a message to stop the stream
        actor_args = {
            'command': 'stop',
            'uri': uri,
            'edits': ls.edits,
            'doc': doc_source,
        }
        for actor in ls.actors:
            ls.tell_actor(actor, actor_args)

    # Add `command.stop` as a "Code Action" too (accessible from the dropdown
    # menu, eg `M-'`.
    server.add_code_action(code_action_stop)
    return server


def code_action_stop(params: CodeActionParams):
    text_document = params.text_document
    return CodeAction(
                title='Stop Streaming Things',
                kind=CodeActionKind.Refactor,
                command=Command(
                    title='Stop Streaming Things',
                    command='command.stop',
                    arguments=[text_document]
                )
            )



'''

Connection to OpenAI API.

'''


from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Range,
    TextDocumentIdentifier,
    WorkspaceEdit,
)
from concurrent.futures import ThreadPoolExecutor
import openai
from threading import Event
from thespian.actors import Actor
import argparse
import logging

from uniteai.edit import init_block, cleanup_block, BlockJob
from uniteai.common import extract_range, find_block, mk_logger, get_nested
from uniteai.server import Server


##################################################
# OpenAI

START_TAG = ':START_OPENAI:'
END_TAG = ':END_OPENAI:'
NAME = 'openai'
log = mk_logger(NAME, logging.WARN)

# A sentinel to be used by LSP clients when calling commands. If a CodeAction
# is called (from a dropdown in the text editor), params are defined in the
# python function here. If a Command is called (IE a key sequence bound to a
# function), the editor provides the values. But if we want the client to defer
# to the LSP server's config, use this sentinel. See usage example in `docs/`.
FROM_CONFIG = 'FROM_CONFIG'
FROM_CONFIG_CHAT = 'FROM_CONFIG_CHAT'
FROM_CONFIG_COMPLETION = 'FROM_CONFIG_COMPLETION'


class OpenAIActor(Actor):
    def __init__(self):
        log.debug('ACTOR INIT')
        self.is_running = False
        self.executor = ThreadPoolExecutor(max_workers=3)
        self.current_future = None
        self.should_stop = Event()
        self.tags = [START_TAG, END_TAG]

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        doc = msg.get('doc')
        edits = msg.get('edits')
        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
is_running: {self.is_running}
should_stop: {self.should_stop.is_set()}
current_future: {self.current_future}

EDITS STATE:
job_thread alive: {edits.job_thread.is_alive() if edits and edits.job_thread else "NOT STARTED"}
%%%%%%%%%%''')
        if command == 'start':
            uri = msg.get('uri')
            range = msg.get('range')
            prompt = msg.get('prompt')
            engine = msg.get('engine')
            max_length = msg.get('max_length')
            edits = msg.get('edits')

            # check if block already exists
            start_ixs, end_ixs = find_block(START_TAG,
                                            END_TAG,
                                            doc)

            if not (start_ixs and end_ixs):
                init_block(NAME, self.tags, uri, range, edits)

            self.start(uri, range, prompt, engine, max_length, edits)

        elif command == 'stop':
            self.stop()

    def start(self, uri, range, prompt, engine, max_length, edits):
        if self.is_running:
            log.info('WARN: ON_START_BUT_RUNNING')
            return
        log.debug('ACTOR START')

        self.is_running = True
        self.should_stop.clear()

        def f(uri_, prompt_, engine_, max_length_, should_stop_, edits_):
            ''' Compose the streaming fn with some cleanup. '''
            openai_stream_fn(uri_, prompt_, engine_, max_length_,
                             should_stop_, edits_)

            # Cleanup
            log.debug('CLEANING UP')
            cleanup_block(NAME, self.tags, uri_, edits_)
            self.is_running = False
            self.current_future = None
            self.should_stop.clear()

        self.current_future = self.executor.submit(
            f, uri, prompt, engine, max_length, self.should_stop, edits
        )
        log.debug('START CAN RETURN')

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running:
            log.info('WARN: ON_STOP_BUT_STOPPED')

        self.should_stop.set()

        if self.current_future:
            self.current_future.result()  # block, wait to finish
            self.current_future = None
        log.debug('FINALLY STOPPED')


##################################################
# OpenAI

COMPLETION_ENGINES = [
    "text-davinci-003",
    "text-davinci-002",
    "ada",
    "babbage",
    "curie",
    "davinci",
]

CHAT_ENGINES = [
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0613",
    "gpt-4",
]


def openai_autocomplete(engine, text, max_length):
    ''' Stream responses from OpenAI's API as a generator. '''
    if engine in COMPLETION_ENGINES:
        response = openai.Completion.create(
          engine=engine,
          prompt=text,
          max_tokens=max_length,
          stream=True
        )
        for message in response:
            generated_text = message['choices'][0]['text']
            yield generated_text
    elif engine in CHAT_ENGINES:
        response = openai.ChatCompletion.create(
          model=engine,
          messages=[{"role": "user", "content": text}],
          stream=True
        )
        for message in response:
            # different json structure than completion endpoint
            delta = message['choices'][0]['delta']
            if 'content' in delta:
                generated_text = delta['content']
                yield generated_text

def openai_stream_fn(uri, prompt, engine, max_length, stop_event, edits):
    log.debug(f'START: OPENAI_STREAM_FN, max_length={max_length}')
    try:
        # Stream the results to LSP Client
        running_text = ''
        for new_text in openai_autocomplete(engine, prompt, max_length):
            # For breaking out early
            if stop_event.is_set():
                log.debug('STREAM_FN received STOP EVENT')
                break
            log.debug(f'NEW: {new_text}')
            # ignore empty strings
            if len(new_text) == 0:
                continue

            running_text += new_text
            job = BlockJob(
                uri=uri,
                start_tag=START_TAG,
                end_tag=END_TAG,
                text=f'\n{running_text}\n',
                strict=False,
            )
            edits.add_job(NAME, job)

        # Streaming is done, and those added jobs were all non-strict. Let's
        # make sure to have one final strict job. Streaming jobs are ok to be
        # dropped, but we need to make sure it does finalize, eg before a
        # strict delete-tags job is added.
        job = BlockJob(
            uri=uri,
            start_tag=START_TAG,
            end_tag=END_TAG,
            text=f'\n{running_text}\n',
            strict=True,
        )
        edits.add_job(NAME, job)
        log.debug('STREAM COMPLETE')
    except Exception as e:
        log.error(f'Error: OpenAI, {e}')


def code_action_gpt(engine, max_length, params: CodeActionParams):
    '''Trigger a GPT Autocompletion response. A code action calls a command,
    which is set up below to `tell` the actor to start streaming a response.'''
    text_document = params.text_document
    range = params.range
    return CodeAction(
        title='OpenAI GPT',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='OpenAI GPT',
            command='command.openaiAutocompleteStream',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, range, engine, max_length]
        )
    )


def code_action_chat_gpt(engine, max_length, params: CodeActionParams):
    '''Trigger a ChatGPT response. A code action calls a command, which is set
    up below to `tell` the actor to start streaming a response. '''
    text_document = params.text_document
    range = params.range
    return CodeAction(
        title='OpenAI ChatGPT',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='OpenAI ChatGPT',
            command='command.openaiAutocompleteStream',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, range, engine, max_length]
        )
    )


##################################################
# Setup

def configure(config_yaml):
    parser = argparse.ArgumentParser()
    parser.add_argument('--openai_completion_engine', default=get_nested(config_yaml, ['openai', 'completion_engine']))
    parser.add_argument('--openai_chat_engine', default=get_nested(config_yaml, ['openai', 'chat_engine']))
    parser.add_argument('--openai_max_length', default=get_nested(config_yaml, ['openai', 'max_length']))
    parser.add_argument('--openai_api_key', default=get_nested(config_yaml, ['openai', 'api_key']))

    # bc this is only concerned with openai params, do not error if extra params
    # are sent via cli.
    args, _ = parser.parse_known_args()
    return args




def initialize(config, server):
    # Config
    openai_chat_engine = config.openai_chat_engine
    openai_completion_engine = config.openai_completion_engine
    openai_max_length = config.openai_max_length
    openai.api_key = config.openai_api_key  # make library aware of api key

    # Actor
    server.add_actor(NAME, OpenAIActor)

    # CodeActions
    server.add_code_action(
        lambda params:
        code_action_gpt(openai_completion_engine, openai_max_length, params))
    server.add_code_action(
        lambda params:
        code_action_chat_gpt(openai_chat_engine, openai_max_length, params))

    # Modify Server
    @server.thread()
    @server.command('command.openaiAutocompleteStream')
    def openai_autocomplete_stream(ls: Server, args):
        if len(args) != 4:
            log.error(f'command.openaiAutocompleteStream: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        range = ls.converter.structure(args[1], Range)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Determine engine, by checking for sentinel values to allow LSP client
        # to defer arguments to server's configuration.
        if args[2] == FROM_CONFIG_CHAT:
            engine = openai_chat_engine
        elif args[2] == FROM_CONFIG_COMPLETION:
            engine = openai_completion_engine
        else:
            engine = args[2]

        # Max Length
        if args[3] == FROM_CONFIG:
            max_length = openai_max_length
        else:
            max_length = args[3]

        # Extract the highlighted region
        prompt = extract_range(doc_source, range)

        # Send a message to start the stream
        actor_args = {
            'command': 'start',
            'uri': uri,
            'range': range,
            'prompt': prompt,
            'engine': engine,
            'max_length': max_length,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)

        # Return null-edit immediately (the rest will stream)
        return WorkspaceEdit()
